from conftest import create_org_app

from onegov.core.security import Secret
from onegov.org.app import OrgApp
from onegov.org.forms import ExportForm
from onegov.org.models import Export
from onegov_testing import Client as BaseClient


class Client(BaseClient):
    skip_first_form = True


def test_export(request):

    class App(OrgApp):
        pass

    @App.export(
        id='my-export',
        form_class=ExportForm,
        permission=Secret,
        title="My Export",
        explanation="Foo bar.")
    class MyExport(Export):
        def run(self, form, session):
            return [[
                ('id', 'foo'),
                ('content', 'bar')
            ]]

    app = create_org_app(request, use_elasticsearch=False, cls=App)

    # the export should work for admins
    admin = Client(app)
    admin.login_admin()

    page = admin.get('/exports')
    assert "My Export" in page
    assert "Foo bar." in page

    page = page.click("My Export")
    page.form['file_format'] = 'json'
    result = page.form.submit().json

    assert result == [
        {
            'id': 'foo',
            'content': 'bar'
        }
    ]

    # but because of our permission editors don't have a chance
    editor = Client(app)
    editor.login_editor()

    page = editor.get('/exports')
    assert "My Export" not in page
    assert "Foo bar." not in page

    page = editor.get('/export/my-export', status=403)
    assert "Ihnen fehlt die nötige Berechtigung" in page



import morepath

from libres.modules.errors import LibresError
from onegov.core.security import Public, Private
from onegov.reservation import Allocation, Resource, ResourceCollection
from onegov.org import OrgApp, utils, _
from onegov.org.elements import Link
from onegov.org.forms import (
    DaypassAllocationForm,
    DaypassAllocationEditForm,
    RoomAllocationForm,
    RoomAllocationEditForm
)
from onegov.org.layout import ResourceLayout, AllocationEditFormLayout
from sqlalchemy.orm import defer, defaultload
from purl import URL


@OrgApp.json(model=Resource, name='slots', permission=Public)
def view_allocations_json(self, request):
    """ Returns the allocations in a fullcalendar compatible events feed.

    See `<http://fullcalendar.io/docs/event_data/events_json_feed/>`_ for
    more information.

    """

    start, end = utils.parse_fullcalendar_request(request, self.timezone)

    if not (start and end):
        return tuple()

    # get all allocations (including mirrors), for the availability calculation
    query = self.scheduler.allocations_in_range(start, end, masters_only=False)
    query = query.order_by(Allocation._start)
    query = query.options(defer(Allocation.data))
    query = query.options(defer(Allocation.group))
    query = query.options(
        defaultload('reserved_slots')
        .defer('reservation_token')
        .defer('allocation_id')
        .defer('end'))

    allocations = query.all()

    # but only return the master allocations
    return tuple(
        e.as_dict() for e in utils.AllocationEventInfo.from_allocations(
            request, self.scheduler, allocations
        )
    )


def get_new_allocation_form_class(resource, request):
    """ Returns the form class for new allocations (different resources have
    different allocation forms).

    """

    if resource.type == 'daypass':
        return DaypassAllocationForm

    if resource.type == 'room':
        return RoomAllocationForm

    raise NotImplementedError


def get_edit_allocation_form_class(allocation, request):
    """ Returns the form class for existing allocations (different resources
    have different allocation forms).

    """

    resource = ResourceCollection(
        request.app.libres_context).by_id(allocation.resource)

    if resource.type == 'daypass':
        return DaypassAllocationEditForm

    if resource.type == 'room':
        return RoomAllocationEditForm

    raise NotImplementedError


@OrgApp.form(model=Resource, template='form.pt', name='new-allocation',
             permission=Private, form=get_new_allocation_form_class)
def handle_new_allocation(self, request, form):
    """ Handles new allocations for differing form classes. """

    if form.submitted(request):
        try:
            allocations = self.scheduler.allocate(
                dates=form.dates,
                whole_day=form.whole_day,
                quota=form.quota,
                quota_limit=form.quota_limit,
                data=form.data,
                partly_available=form.partly_available
            )
        except LibresError as e:
            utils.show_libres_error(e, request)
        else:
            request.success(_("Successfully added ${n} allocations", mapping={
                'n': len(allocations)
            }))

            self.highlight_allocations(allocations)
            return morepath.redirect(request.link(self))
    elif not request.POST:
        start, end = utils.parse_fullcalendar_request(request, self.timezone)
        whole_day = request.params.get('whole_day') == 'yes'

        if start and end:
            if whole_day:
                form.start.data = start
                form.end.data = end

                if hasattr(form, 'as_whole_day'):
                    form.as_whole_day.data = 'yes'
            else:
                form.start.data = start
                form.end.data = end

                if hasattr(form, 'as_whole_day'):
                    form.as_whole_day.data = 'no'

                if hasattr(form, 'start_time'):
                    form.start_time.data = start
                    form.end_time.data = end

    layout = ResourceLayout(self, request)
    layout.breadcrumbs.append(Link(_("New allocation"), '#'))
    layout.editbar_links = None

    return {
        'layout': layout,
        'title': _("New allocation"),
        'form': form
    }


@OrgApp.form(model=Allocation, template='form.pt', name='edit',
             permission=Private, form=get_edit_allocation_form_class)
def handle_edit_allocation(self, request, form):
    """ Handles edit allocation for differing form classes. """

    resources = ResourceCollection(request.app.libres_context)
    resource = resources.by_id(self.resource)

    # this is a bit of a hack to keep the current view when a user drags an
    # allocation around, which opens this form and later leads back to the
    # calendar - if the user does this on the day view we want to return to
    # the same day view after the process
    # therefore we set the view on the resource (where this is okay) and on
    # the form action (where it's a bit of a hack), to ensure that the view
    # parameter is around for the whole time
    if 'view' in request.params:
        resource.view = view = request.params['view']
        form.action = URL(form.action).query_param('view', view).as_string()

    if form.submitted(request):
        new_start, new_end = form.dates

        try:
            resource.scheduler.move_allocation(
                master_id=self.id,
                new_start=new_start,
                new_end=new_end,
                new_quota=form.quota,
                quota_limit=form.quota_limit
            )
        except LibresError as e:
            utils.show_libres_error(e, request)
        else:
            request.success(_("Your changes were saved"))
            resource.highlight_allocations([self])

            return morepath.redirect(request.link(resource))
    elif not request.POST:
        form.apply_model(self)

        start, end = utils.parse_fullcalendar_request(request, self.timezone)
        if start and end:
            form.apply_dates(start, end)

    return {
        'layout': AllocationEditFormLayout(self, request),
        'title': _("Edit allocation"),
        'form': form
    }


@OrgApp.view(model=Allocation, request_method='DELETE', permission=Private)
def handle_delete_allocation(self, request):
    """ Deletes the given resource (throwing an error if there are existing
    reservations associated with it).

    """
    request.assert_valid_csrf_token()

    resource = request.app.libres_resources.by_allocation(self)
    resource.scheduler.remove_allocation(id=self.id)

    @request.after
    def trigger_calendar_update(response):
        response.headers.add('X-IC-Trigger', 'rc-allocations-changed')



from setuptools import setup, find_packages

name = 'onegov.org'
description = (
    'A OneGov Cloud base application for organisations.'
)
version = '0.41.0'


def get_long_description():
    readme = open('README.rst').read()
    history = open('HISTORY.rst').read()

    # cut the part before the description to avoid repetition on pypi
    readme = readme[readme.index(description) + len(description):]

    return '\n'.join((readme, history))


setup(
    name=name,
    version=version,
    description=description,
    long_description=get_long_description(),
    url='http://github.com/OneGov/onegov.org',
    author='Seantis GmbH',
    author_email='info@seantis.ch',
    license='GPLv2',
    packages=find_packages(exclude=['ez_setup']),
    namespace_packages=name.split('.')[:-1],
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    python_requires='>=3.6',
    install_requires=[
        'babel',
        'click',
        'isodate',
        'lxml',
        'onegov.chat',
        'onegov.core>=0.41.0',
        'onegov.directory>=0.4.1',
        'onegov.event>=0.9.0',
        'onegov.file>=0.2.1',
        'onegov.form>=0.18.0',
        'onegov.foundation>=0.0.7',
        'onegov.gis',
        'onegov.newsletter>=0.1.0',
        'onegov.page>=0.1.0',
        'onegov.pay',
        'onegov.people>=0.6.0',
        'onegov.recipient',
        'onegov.reservation>=0.3.0',
        'onegov.search>=1.0.0',
        'onegov.shared',
        'onegov.ticket>=0.4.1',
        'onegov.user>=0.11.0',
        'pillow',
        'purl',
        'python-dateutil',
        'python-magic>=0.4.12',
        'pytz',
        'PyYAML',
        'rjsmin',
        'sedate',
        'translationstring',
        'webtest',
    ],
    extras_require=dict(
        test=[
            'coverage',
            'freezegun',
            'onegov_testing',
            'pyquery',
            'pytest-localserver',
            'requests-mock',
            'vcrpy',
        ],
    ),
    entry_points={
        'onegov': [
            'upgrade = onegov.org.upgrade'
        ],
        'console_scripts': [
            'onegov-org=onegov.org.cli:cli'
        ]
    },
    classifiers=[
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
    ]
)



from collections import OrderedDict
from onegov.core.orm.types import UUID
from onegov.core.security import Private
from onegov.form import FormDefinition
from onegov.form import FormRegistrationWindow
from onegov.form import FormSubmission
from onegov.form import FormSubmissionCollection
from onegov.org import _
from onegov.org import OrgApp
from onegov.org.forms import FormSubmissionsExport
from onegov.org.layout import FormSubmissionLayout
from onegov.org.new_elements import Link
from onegov.org.utils import keywords_first
from onegov.ticket import Ticket
from sedate import align_range_to_day, as_datetime, standardize_date
from sqlalchemy import and_


@OrgApp.form(
    model=FormDefinition,
    name='export',
    permission=Private,
    form=FormSubmissionsExport,
    template='export.pt')
def handle_form_submissions_export(self, request, form):

    layout = FormSubmissionLayout(self, request)
    layout.breadcrumbs.append(Link(_("Export"), '#'))
    layout.editbar_links = None

    if form.submitted(request):
        submissions = FormSubmissionCollection(request.session, name=self.name)

        if form.data['selection'] == 'date':
            subset = subset_by_date(
                submissions=submissions,
                start=form.data['start'],
                end=form.data['end'],
                timezone=layout.timezone)

        elif form.data['selection'] == 'window':
            subset = subset_by_window(
                submissions=submissions,
                window_ids=form.data['registration_window'])
        else:
            raise NotImplementedError

        subset = configure_subset(subset)

        field_order, results = run_export(
            subset=subset,
            nested=form.format == 'json',
            formatter=layout.export_formatter(form.format))

        return form.as_export_response(results, self.title, key=field_order)

    return {
        'title': _("Export"),
        'layout': layout,
        'form': form,
        'explanation': _("Exports the submissions of the given date range.")
    }


def subset_by_date(submissions, start, end, timezone):
    start, end = align_range_to_day(
        standardize_date(as_datetime(start), timezone),
        standardize_date(as_datetime(end), timezone),
        timezone)

    subset = submissions.query().filter_by(state='complete').filter(and_(
        FormSubmission.received >= start,
        FormSubmission.received <= end))

    return subset


def subset_by_window(submissions, window_ids):
    subset = submissions.query().filter_by(state='complete').filter(
        FormSubmission.registration_window_id.in_(window_ids))

    return subset


def configure_subset(subset):
    subset = subset.join(
        Ticket,
        FormSubmission.id == Ticket.handler_id.cast(UUID)
    )

    subset = subset.join(
        FormRegistrationWindow,
        FormSubmission.registration_window_id == FormRegistrationWindow.id,
        isouter=True
    )

    subset = subset.with_entities(
        FormSubmission.title,
        FormSubmission.email,
        FormSubmission.received,
        FormSubmission.payment_method,
        FormSubmission.data,
        Ticket.number.label('ticket_number'),
        FormSubmission.registration_state,
        FormRegistrationWindow.start.label('registration_window_start'),
        FormRegistrationWindow.end.label('registration_window_end'),
    )

    return subset


def run_export(subset, nested, formatter):
    keywords = (
        'ticket_number',
        'title',
        'email',
        'received',
        'payment_method',
        'registration_state',
        'registration_window_start',
        'registration_window_end'
    )

    def transform(submission):
        r = OrderedDict()

        for keyword in keywords:
            r[keyword] = formatter(getattr(submission, keyword))

        if nested:
            r['form'] = {
                k: formatter(v)
                for k, v in submission.data.items()
            }
        else:
            for k, v in submission.data.items():
                r[f'form_{k}'] = formatter(v)

        return r

    return keywords_first(keywords), tuple(transform(s) for s in subset)



from collections import OrderedDict
from itertools import groupby
from unidecode import unidecode


class AtoZ(object):

    def __init__(self, request):
        self.request = request

    def sortkey(self, item):
        return unidecode(self.get_title(item)[0].upper())

    def get_items_by_letter(self):
        items_by_letter = OrderedDict()

        for letter, items in groupby(self.get_items(), self.sortkey):
            items_by_letter[unidecode(letter)] = tuple(
                sorted(items, key=lambda i: unidecode(self.get_title(i)))
            )

        return items_by_letter

    def get_title(self, item):
        raise NotImplementedError

    def get_items(self):
        raise NotImplementedError
